from django.db import router
from django.db.models.deletion import Collector
from django.urls import NoReverseMatch, reverse


def get_deleted_objects(root):
    """Collect linked objects required to be deleted.

    And maps each one to the `detail` page link.
    """
    using = router.db_for_write(root)
    collector = Collector(using=using)
    collector.collect([root])

    def format_callback(obj):
        opts = obj._meta

        try:
            return reverse(
                '{}:{}_{}_details'.format(opts.app_label, opts.app_label, opts.model_name),
                args=[obj._get_pk_val()])
        except NoReverseMatch:
            return None

    to_delete = collector.nested(format_callback)
    return to_delete



from django.http import HttpResponseRedirect

try:
    from urllib.parse import urlencode, parse_qs, urlsplit, urlunsplit
except ImportError:
    from urlparse import parse_qs, urlsplit, urlunsplit
    from urllib import urlencode


class SmoothNavigationMiddleware(object):
    """Keep `?back=` queryset parameter on POST requests."""
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):  # noqa D102
        response = self.get_response(request)

        if isinstance(response, HttpResponseRedirect):
            back = request.GET.get('back')
            if back:
                _, _, back_path, _, _ = urlsplit(back)
                scheme, netloc, path, query_string, fragment = urlsplit(response['location'])
                query_params = parse_qs(query_string)

                if path == back_path:
                    query_params.pop('back', None)
                elif 'back' not in query_params:
                    query_params['back'] = [back]

                new_query_string = urlencode(query_params, doseq=True)
                response['location'] = urlunsplit((scheme, netloc, path, new_query_string, fragment))

        return response



import errno
import io
import shutil
import os
from importlib import import_module
from os import path

import django
from django.conf import settings
from django.template import Context, Engine
from django.core.management.base import CommandError
from django.core.management.utils import handle_extensions
from django.core.management.templates import TemplateCommand
from django.utils import six

import material


class Command(TemplateCommand):
    help = (
        "Create a Frontend Module directory structure for the given app name in "
        "the current directory or optionally in the given directory."
    )
    missing_args_message = "You must provide an application name."
    rewrite_template_suffixes = (
        ('.py-tpl', '.py'),
        ('.html-tpl', '.html'),
    )

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.set_defaults(extensions=['py', 'html'])

    def handle_template(self, template, subdir):
        """
        Determine where the app or project templates are.

        Use django.__path__[0] as the default because we don't
        know into which directory Django has been installed.
        """
        if template is None:
            return path.join(material.__path__[0], 'conf', subdir)
        return super(Command, self).handle_template(template, subdir)

    def _validate_name(self, app_name):
        super(Command, self).validate_name(app_name, "module")
        try:
            import_module(app_name)
        except ImportError:
            pass
        else:
            raise CommandError(
                "%r conflicts with the name of an existing Python module and "
                "cannot be used as an app name. Please try another name." % app_name
            )

    def _expand_target_dir(self, app_name, target):
        if target is None:
            top_dir = path.join(os.getcwd(), app_name)
            try:
                os.makedirs(top_dir)
            except OSError as e:
                if e.errno == errno.EEXIST:
                    message = "'%s' already exists" % top_dir
                else:
                    message = e
                raise CommandError(message)
        else:
            top_dir = os.path.abspath(path.expanduser(target))
            if not os.path.exists(top_dir):
                raise CommandError("Destination directory '%s' does not "
                                   "exist, please create it first." % top_dir)
        return top_dir

    def handle(self, **options):
        self.verbosity = options['verbosity']
        app_name, target = options.pop('name'), options.pop('directory')
        extensions = tuple(handle_extensions(options['extensions']))
        extra_files = []
        for file in options['files']:
            extra_files.extend(map(lambda x: x.strip(), file.split(',')))

        self._validate_name(app_name)

        self.paths_to_remove = []
        top_dir = self._expand_target_dir(app_name, target)

        camel_case_app_name = ''.join(x for x in app_name.title() if x != '_')
        context = Context(dict(options, **{
            'app_name': app_name,
            'camel_case_app_name': camel_case_app_name,
            'base_directory': top_dir,
            'unicode_literals': '' if six.PY3 else '# -*- coding: utf-8 -*-\n'
                                                   'from __future__ import unicode_literals\n\n',
        }), autoescape=False)

        # Setup a stub settings environment for template rendering
        if not settings.configured:
            settings.configure()
            django.setup()

        template_dir = self.handle_template(options['template'], 'module_template')
        prefix_length = len(template_dir) + 1

        for root, dirs, files in os.walk(template_dir):
            path_rest = root[prefix_length:]
            relative_dir = path_rest.replace('module_name', app_name)
            if relative_dir:
                target_dir = path.join(top_dir, relative_dir)
                if not path.exists(target_dir):
                    os.mkdir(target_dir)

            for filename in files:
                old_path = path.join(root, filename)
                new_path = path.join(top_dir, relative_dir, filename.replace('module_name', app_name))

                for old_suffix, new_suffix in self.rewrite_template_suffixes:
                    if new_path.endswith(old_suffix):
                        new_path = new_path[:-len(old_suffix)] + new_suffix
                        break  # Only rewrite once

                if path.exists(new_path):
                    raise CommandError(
                        "{} already exists, overlaying a project or app into an existing "
                        "directory won't replace conflicting files".format(new_path))

                if new_path.endswith(extensions) or filename in extra_files:
                    with io.open(old_path, 'r', encoding='utf-8') as template_file:
                        content = template_file.read()
                    template = Engine().from_string(content)
                    content = template.render(context)
                    with io.open(new_path, 'w', encoding='utf-8') as new_file:
                        new_file.write(content)
                else:
                    shutil.copyfile(old_path, new_path)

                if self.verbosity >= 2:
                    self.stdout.write("Creating %s\n" % new_path)
                try:
                    shutil.copymode(old_path, new_path)
                    self.make_writeable(new_path)
                except OSError:
                    self.stderr.write(
                        "Notice: Couldn't set permission bits on %s. You're "
                        "probably using an uncommon filesystem setup. No "
                        "problem." % new_path, self.style.NOTICE)

        if self.paths_to_remove:
            if self.verbosity >= 2:
                self.stdout.write("Cleaning up temporary files.\n")
            for path_to_remove in self.paths_to_remove:
                if path.isfile(path_to_remove):
                    os.remove(path_to_remove)
                else:
                    shutil.rmtree(path_to_remove)

        if self.verbosity >= 1:
            self.stdout.write("{} module created\n" .format(camel_case_app_name))
            self.stdout.write(
                "To enable the module you need to add '{}.apps.{}Config' "
                "to the INSTALLED_APPS setting".format(app_name, camel_case_app_name))
            self.stdout.write('and run ./manage.py migrate after that\n')



from django.apps import AppConfig
from django.contrib import admin
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from ..frontend.apps import ModuleMixin
from ..frontend.urlconf import ModuleURLResolver


class MaterialAdminConfig(ModuleMixin, AppConfig):
    """Material Admin frontend module."""

    name = 'material.admin'
    label = "material_admin"

    icon = '<i class="material-icons">build</i>'
    verbose_name = _("Administration")
    order = 1000

    @property
    def urls(self):  # noqa D102
        return ModuleURLResolver(r'^admin/', admin.site.urls[0], namespace='admin', module=self)

    def index_url(self):  # noqa D102
        return reverse('admin:index'.format(self.label))

    def has_perm(self, user):  # noqa D102
        return user.is_staff



import re

from django import forms


ORDER_RE = re.compile(r'datatable-order\[(\d+)\]\[(\w+)\]')


class DatatableRequestForm(forms.Form):
    """Sanitize  incoming datatables request."""

    draw = forms.IntegerField()
    start = forms.IntegerField()
    length = forms.IntegerField()

    def clean_ordering(self):
        order = {}
        for key, value in self.data.items():
            if key.startswith('datatable-order['):
                match = ORDER_RE.match(key)
                pos, attr = match.groups()
                if pos not in order:
                    order[pos] = {}
                order[pos][attr] = value

        self.cleaned_data['ordering'] = [spec[1] for spec in sorted(order.items())]

    def clean(self):
        self.clean_ordering()
        return self.cleaned_data



from __future__ import unicode_literals

from django.conf.urls import url
from django.contrib.auth import get_permission_codename

from .create import CreateModelView
from .delete import DeleteModelView
from .detail import DetailModelView
from .list import ListModelView
from .update import UpdateModelView


DEFAULT = object()


class BaseViewset(object):
    """Base router-like class for frontend viewset."""

    @property
    def urls(self):
        """Collect url specs from the instance attributes.

        Assumes that each attribute with name ending with `_view`
        contains triple (regexp, view, url_name)
        """
        result = []

        format_kwargs = {
            'model_name': self.model._meta.model_name
        }

        url_entries = (
            getattr(self, attr)
            for attr in dir(self)
            if attr.endswith('_view')
            if isinstance(getattr(self, attr), (list, tuple))
        )
        for url_entry in url_entries:
            regexp, view, name = url_entry
            result.append(
                url(regexp.format(**format_kwargs),
                    view,
                    name=name.format(**format_kwargs))
            )

        return result


class ModelViewSet(BaseViewset):
    """Model Create/Read/Update/Delete/List viewset.

    Lightweight alternative for django admin. CRUD interface based in
    the django generic views. Viewset provides a simple place to
    configure the views and share the common configuration parameters.

    :keyword model: Model class views

    :keyword queryset: Base views queryset

    :keyword list_display: List of fields for ListView

    :keyword list_display_links: List of fields form `list_display`
                                 linked to change view

    :keyword ordering: Defaul ListView ordering

    :keyword create_view_class: CBV for create an object

    :keyword list_view_class:  CBV for create to list objects

    :keyword detail_view_class:  CBV to show an object detail

    :keyword update_view_class: CBV to update an object

    :keyword delete_view_class: CBV to delete an object

    :keyword layout: Layout for django-material forms

    :keyword form_class: Form Class for Create and Update views

    :keyword form_widgets: Custom widgets for the model form
                           in the Create and Update views, if
                           no custom form_class provided.

    There is no specifically requirements to CBV, by overriding
    corresponding `get_<op>_view` method you can even use function
    based views with this viewset.

    Example::

        class CityViewSet(ModelViewSet):
            model = models.Sea
            list_display = ('name', 'parent', 'ocean', 'sea_area', )
            layout = Layout(
                Row('name', 'parent'),
                'ocean',
                Row('area', 'avg_depth', 'max_depth'),
                'basin_countries'
            )

            def sea_area(self, sea):
                return None if sea.area == 0 else sea.area

    """

    model = None
    queryset = DEFAULT
    list_display = DEFAULT
    list_display_links = DEFAULT
    ordering = DEFAULT

    layout = DEFAULT
    form_class = DEFAULT
    form_widgets = DEFAULT

    def filter_kwargs(self, view_class, **kwargs):
        """Add defaults and filter kwargs to only those that view can accept.

        Viewset pass to the view only kwargs that have non DEFAULT values,
        and if the view have a corresponding attribute.

        In addition, if view has `viewset` attribute, it will receive
        the `self` instance

        """
        result = {
            'model': self.model,
            'viewset': self,
            'queryset': self.queryset,
        }
        result.update(kwargs)
        return {name: value for name, value in result.items()
                if hasattr(view_class, name)
                if value is not DEFAULT}

    """
    Create
    """
    create_view_class = CreateModelView

    def get_create_view(self):
        """Function view for create an object."""
        return self.create_view_class.as_view(**self.get_create_view_kwargs())

    def get_create_view_kwargs(self, **kwargs):
        """Configuration arguments for create view.

        May not be called if `get_create_view` is overridden.
        """
        result = {
            'layout': self.layout,
            'form_class': self.form_class,
            'form_widgets': self.form_widgets,
        }
        result.update(kwargs)
        return self.filter_kwargs(self.create_view_class, **result)

    @property
    def create_view(self):
        """Triple (regexp, view, name) for create view url config."""
        return [
            r'^add/$',
            self.get_create_view(),
            '{model_name}_add',
        ]

    def has_add_permission(self, request):
        """Default add permission check for a detail and list views.

        May not be called if views have own implementation.
        """
        opts = self.model._meta
        codename = get_permission_codename('add', opts)
        return request.user.has_perm('{}.{}'.format(opts.app_label, codename))

    """
    Detail
    """
    detail_view_class = DetailModelView

    def get_detail_view(self):
        """Function view for object detail."""
        return self.detail_view_class.as_view(**self.get_detail_view_kwargs())

    def get_detail_view_kwargs(self, **kwargs):
        """Configuration arguments for detail view.

        May not be called if `get_detail_view` is overridden.
        """
        return self.filter_kwargs(self.detail_view_class, **kwargs)

    def has_view_permission(self, request, obj=None):
        """Default view permission check for a detail and list views.

        May not be called if views have own implementation.
        """
        opts = self.model._meta
        codename = get_permission_codename('view', opts)
        view_perm = '{}.{}'.format(opts.app_label, codename)
        if request.user.has_perm(view_perm):
            return True
        elif request.user.has_perm(view_perm, obj=obj):
            return True
        return self.has_change_permission(request, obj=obj)

    @property
    def detail_view(self):
        """Triple (regexp, view, name) for detail view url config."""
        return [
            r'^(?P<pk>.+)/detail/$',
            self.get_detail_view(),
            '{model_name}_detail'
        ]

    """
    List
    """
    list_view_class = ListModelView

    def get_list_view(self):
        """Function view for objects list."""
        return self.list_view_class.as_view(**self.get_list_view_kwargs())

    def get_list_view_kwargs(self, **kwargs):
        """Configuration arguments for list view.

        May not be called if `get_list_view` is overridden.
        """
        result = {
            'list_display': self.list_display,
            'list_display_links': self.list_display_links,
            'ordering': self.ordering
        }
        result.update(kwargs)
        return self.filter_kwargs(self.list_view_class, **result)

    @property
    def list_view(self):
        """Triple (regexp, view, name) for list view url config."""
        return [
            '^$',
            self.get_list_view(),
            '{model_name}_list'
        ]

    """
    Update
    """
    update_view_class = UpdateModelView

    def get_update_view(self):
        """Function view for update an object."""
        return self.update_view_class.as_view(**self.get_update_view_kwargs())

    def get_update_view_kwargs(self, **kwargs):
        """Configuration arguments for update view.

        May not be called if `get_update_view` is overridden.
        """
        result = {
            'layout': self.layout,
            'form_class': self.form_class,
            'form_widgets': self.form_widgets,
        }
        result.update(kwargs)
        return self.filter_kwargs(self.update_view_class, **result)

    def has_change_permission(self, request, obj=None):
        """Default change permission check for a update view.

        May not be called if update view have own implementation.
        """
        opts = self.model._meta
        codename = get_permission_codename('change', opts)
        change_perm = '{}.{}'.format(opts.app_label, codename)
        if request.user.has_perm(change_perm):
            return True
        return request.user.has_perm(change_perm, obj=obj)

    @property
    def update_view(self):
        """Triple (regexp, view, name) for update view url config."""
        return [
            r'^(?P<pk>.+)/change/$',
            self.get_update_view(),
            '{model_name}_change',
        ]

    """
    Delete
    """
    delete_view_class = DeleteModelView

    def get_delete_view(self):
        """Function view for delete an object."""
        return self.delete_view_class.as_view(**self.get_delete_view_kwargs())

    def has_delete_permission(self, request, obj=None):
        """Default delete permission check for a delete view.

        May not be called if delete view have own implementation.
        """
        opts = self.model._meta
        codename = get_permission_codename('delete', opts)
        delete_perm = '{}.{}'.format(opts.app_label, codename)
        if request.user.has_perm(delete_perm):
            return True
        return request.user.has_perm(delete_perm, obj=obj)

    def get_delete_view_kwargs(self, **kwargs):
        """Configuration arguments for delete view.

        May not be called if `get_delete_view` is overridden.
        """
        return self.filter_kwargs(self.delete_view_class, **kwargs)

    @property
    def delete_view(self):
        """Triple (regexp, view, name) for delete view url config."""
        return [
            r'^(?P<pk>.+)/delete/$',
            self.get_delete_view(),
            '{model_name}_delete'
        ]



from django.conf import settings

from .urlconf import ModuleURLResolver, frontend_url
from .registry import modules


__all__ = (
    'ModuleURLResolver', 'frontend_url', 'modules'
)

default_app_config = 'material.frontend.apps.MaterialFrontendConfig'


if getattr(settings, 'MATERIAL_FRONTEND_AUTOREGISTER', True):
    # Register middleware
    if 'material.frontend.middleware.SmoothNavigationMiddleware' not in settings.MIDDLEWARE:
        settings.MIDDLEWARE += ('material.frontend.middleware.SmoothNavigationMiddleware',)

    # Context processors
    for engine in settings.TEMPLATES:
        if engine['BACKEND'] == 'django.template.backends.django.DjangoTemplates':
            if 'OPTIONS' not in engine:
                engine['OPTIONS'] = {}
            if 'context_processors' not in engine['OPTIONS']:
                engine['OPTIONS']['context_processors'] = []
            if 'material.frontend.context_processors.modules' not in engine['OPTIONS']['context_processors']:
                engine['OPTIONS']['context_processors'].append('material.frontend.context_processors.modules')



from __future__ import unicode_literals

from django.contrib.auth import get_permission_codename
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import models
from django.http import Http404
from django.urls import reverse
from django.views import generic


class DetailModelView(generic.DetailView):
    """Thin wrapper for `generic.DetailView`."""

    viewset = None

    def get_object_data(self):
        """List of object fields to display.

        Choice fields values are expanded to readable choice label.
        """
        for field in self.object._meta.fields:
            if isinstance(field, models.AutoField):
                continue
            elif field.auto_created:
                continue
            else:
                choice_display_attr = "get_{}_display".format(field.name)
            if hasattr(self.object, choice_display_attr):
                value = getattr(self.object, choice_display_attr)()
            else:
                value = getattr(self.object, field.name)

            if value is not None:
                yield (field.verbose_name.title(), value)

    def has_view_permission(self, request, obj):
        """Object view permission check.

        If view had a `viewset`, the `viewset.has_view_permission` used.
        """
        if self.viewset is not None:
            return self.viewset.has_view_permission(request, obj)

        # default lookup for the django permission
        opts = self.model._meta
        codename = get_permission_codename('view', opts)
        view_perm = '{}.{}'.format(opts.app_label, codename)
        if request.user.has_perm(view_perm):
            return True
        elif request.user.has_perm(view_perm, obj=obj):
            return True
        return self.has_change_permission(request, obj=obj)

    def has_change_permission(self, request, obj):
        """Object chane permission check.

        If view had a `viewset`, the `viewset.has_change_permission` used.

        If true, view will show `Change` link to the Change view.
        """
        if self.viewset is not None:
            return self.viewset.has_change_permission(request, obj)

        # default lookup for the django permission
        opts = self.model._meta
        codename = get_permission_codename('change', opts)
        change_perm = '{}.{}'.format(opts.app_label, codename)
        if request.user.has_perm(change_perm):
            return True
        return request.user.has_perm(change_perm, obj=obj)

    def has_delete_permission(self, request, obj):
        """Object delete permission check.

        If true, view will show `Delete` link to the Delete view.
        """
        if self.viewset is not None:
            return self.viewset.has_delete_permission(request, obj)

        # default lookup for the django permission
        opts = self.model._meta
        codename = get_permission_codename('delete', opts)
        delete_perm = '{}.{}'.format(opts.app_label, codename)
        if request.user.has_perm(delete_perm):
            return True
        return request.user.has_perm(delete_perm, obj=obj)

    def get_object(self):
        """Retrieve the object.

        Check object view permission at the same time.
        """
        queryset = self.get_queryset()
        model = queryset.model
        pk = self.kwargs.get(self.pk_url_kwarg)
        if pk is not None:
            try:
                self.kwargs[self.pk_url_kwarg] = model._meta.pk.to_python(pk)
            except (ValidationError, ValueError):
                raise Http404

        obj = super(DetailModelView, self).get_object()
        if not self.has_view_permission(self.request, obj):
            raise PermissionDenied
        return obj

    def get_context_data(self, **kwargs):
        """Additional context data for detail view.

        :keyword object_data: List of fields and values of the object
        :keyword change_url: Link to the change view
        :keyword delete_url: Link to the delete view
        """
        opts = self.model._meta

        kwargs['object_data'] = self.get_object_data()
        if self.has_change_permission(self.request, self.object):
            kwargs['change_url'] = reverse(
                '{}:{}_change'.format(opts.app_label, opts.model_name),
                args=[self.object.pk])
        if self.has_delete_permission(self.request, self.object):
            kwargs['delete_url'] = reverse(
                '{}:{}_delete'.format(opts.app_label, opts.model_name),
                args=[self.object.pk])

        return super(DetailModelView, self).get_context_data(**kwargs)

    def get_template_names(self):
        """
        List of templates for the view.

        If no `self.template_name` defined, returns::

             [<app_label>/<model_label>_detail.html
              'material/frontend/views/detail.html']
        """
        if self.template_name is None:
            opts = self.model._meta
            return [
                '{}/{}{}.html'.format(
                    opts.app_label,
                    opts.model_name,
                    self.template_name_suffix),
                'material/frontend/views/detail.html',
            ]

        return [self.template_name]



from .create import CreateModelView
from .delete import DeleteModelView
from .detail import DetailModelView
from .list import ListModelView
from .update import UpdateModelView
from .viewset import ModelViewSet


__all__ = (
    'CreateModelView', 'ListModelView', 'UpdateModelView',
    'DeleteModelView', 'DetailModelView', 'ModelViewSet',
)



from __future__ import unicode_literals

from django.contrib.auth import get_permission_codename
from django.core.exceptions import PermissionDenied
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.views import generic

from .mixins import MessageUserMixin, ModelViewMixin


class CreateModelView(MessageUserMixin, ModelViewMixin, generic.CreateView):
    """Thin `generic.CreateView` wrapper plays nice with `ModelViewSet`."""

    template_name_suffix = '_create'

    def has_add_permission(self, request):
        """Object add permission check.

        If view had a `viewset`, the `viewset.has_add_permission` used.
        """
        if self.viewset is not None:
            return self.viewset.has_add_permission(request)

        # default lookup for the django permission
        opts = self.model._meta
        codename = get_permission_codename('add', opts)
        return request.user.has_perm('{}.{}'.format(opts.app_label, codename))

    def get_success_url(self):
        """Redirect back to the detail view if no `success_url` is configured."""
        if self.success_url is None:
            opts = self.model._meta
            return reverse('{}:{}_detail'.format(
                opts.app_label, opts.model_name), args=[self.object.pk]
            )
        return super(ModelViewMixin, self).get_success_url()

    def message_user(self):
        self.success(_('The {name} "{link}" was added successfully.'))

    def dispatch(self, request, *args, **kwargs):
        if not self.has_add_permission(self.request):
            raise PermissionDenied

        return super(CreateModelView, self).dispatch(request, *args, **kwargs)



from django.db import models
from django.core.cache import cache
from django.dispatch import receiver
from django.db.models.signals import post_save
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _


class ModuleManager(models.Manager):
    """Base module state manager."""

    def installed(self, module):
        """By default, all modules considered installed."""
        installed_cache = cache.get('material.modules.installed')
        if not installed_cache:
            installed_cache = [mod.label for mod in self.get_queryset().filter(installed=True)]
            cache.set('material.modules.installed', installed_cache, 60 * 60 * 24)
        return module in installed_cache


@python_2_unicode_compatible
class Module(models.Model):
    """Keep module installed state in the database."""

    label = models.SlugField(_('label'))
    installed = models.BooleanField(_('installed'), default=True)

    objects = ModuleManager()

    class Meta:
        verbose_name = _('module')
        verbose_name_plural = _('modules')

    def __str__(self):
        return self.label


@receiver(post_save, sender=Module)
def _clean_installed_cache(sender, **kwargs):
    cache.delete('material.modules.installed')



"""Django-material - material design for the django framework."""

from .base import (
    Layout, Fieldset, Row, Column, Span, Field,
    Span2, Span3, Span4, Span5, Span6, Span7,
    Span8, Span9, Span10, Span11, Span12,
    LayoutMixin)


__all__ = (
    'Layout', 'Fieldset', 'Row', 'Column', 'Span', 'Field',
    'Span2', 'Span3', 'Span4', 'Span5', 'Span6', 'Span7',
    'Span8', 'Span9', 'Span10', 'Span11', 'Span12',
    'LayoutMixin'
)



import os
import re
from collections import defaultdict

from django.forms.utils import flatatt
from django.forms.forms import BoundField
from django.template import Library
from django.template.base import (
    TemplateSyntaxError, Node, Variable, token_kwargs)
from django.template.loader import get_template
from django.template.loader_tags import IncludeNode
from django.utils.safestring import mark_safe

from ..compat import context_flatten


register = Library()

ATTRS_RE = re.compile(
    r'(?P<attr>[-\w]+)(\s*=\s*[\'"](?P<val>.*?)[\'"])?',
    re.MULTILINE | re.DOTALL
)


def _render_parts(context, parts_list):
    parts = context['form_parts']

    for partnode in parts_list:
        part = partnode.resolve_part(context)
        if partnode.section not in parts[part]:
            value = partnode.render(context)
            parts[part][partnode.section] = value


@register.tag('form')
class FormNode(Node):
    """
    Template based form rendering.

    Example::

        {% form template='material/form.html' form=form layout=view.layout %}
            {% part form.email prepend %}
                <span class="input-group-addon" id="basic-addon1">@</span>
            {% endpart %}
        {% endform %}
    """

    def __init__(self, parser, token):  # noqa D102
        bits = token.split_contents()
        remaining_bits = bits[1:]

        self.kwargs = token_kwargs(remaining_bits, parser)

        if remaining_bits:
            raise TemplateSyntaxError("%r received an invalid token: %r" %
                                      (bits[0], remaining_bits[0]))

        for key in self.kwargs:
            if key not in ('form', 'layout', 'template'):
                raise TemplateSyntaxError("%r received an invalid key: %r" %
                                          (bits[0], key))

            self.kwargs[key] = self.kwargs[key]

        self.nodelist = parser.parse(('end{}'.format(bits[0])))
        parser.delete_first_token()

    def render(self, context):  # noqa D102
        form = self.kwargs.get('form')
        form = form.resolve(context) if form else context.get('form')

        if form is None:
            return ''

        # Take one of view.layout or form.layout
        layout = self.kwargs.get('layout')
        if layout is not None:
            layout = layout.resolve(context)

        if layout is None:
            if 'view' in context:
                view = context['view']
                if hasattr(view, 'layout'):
                    layout = view.layout
        if layout is None:
            if hasattr(form, 'layout'):
                layout = form.layout

        template_name = self.kwargs.get('template')
        if template_name is None:
            template_name = 'material/form.html'
        else:
            template_name = template_name.resolve(context)
        template = get_template(template_name)

        # Render form and parts
        parts = defaultdict(dict)  # part -> section -> value
        attrs = defaultdict(dict)  # field -> section -> atts

        with context.push(
                form=form,
                layout=layout,
                form_template_pack=os.path.dirname(template_name),
                form_parts=parts,
                form_widget_attrs=attrs):

            # direct children
            children = (
                node for node in self.nodelist
                if isinstance(node, FormPartNode)
            )
            _render_parts(context, children)

            attrs = (
                node for node in self.nodelist
                if isinstance(node, WidgetAttrNode)
            )
            for attr in attrs:
                attr.render(context)

            # include
            children = (
                node for node in self.nodelist
                if isinstance(node, IncludeNode)
            )
            for included_list in children:
                included = included_list.template.resolve(context)
                children = (
                    node for node in included.nodelist
                    if isinstance(node, FormPartNode)
                )
                _render_parts(context, children)

                attrs = (
                    node for node in self.nodelist
                    if isinstance(node, WidgetAttrNode)
                )
                for attr in attrs:
                    attr.render(context)

            return template.render(context_flatten(context))


@register.tag('part')
class FormPartNode(Node):
    """Named piece of HTML layout."""

    def __init__(self, parser, token):  # noqa D102
        bits = token.split_contents()

        if len(bits) > 5:
            raise TemplateSyntaxError(
                "%r accepts at most 4 arguments (part_id, section,"
                " asvar, varname), got: {}".format(bits[0], ','.join(bits[1:]))
            )

        self.part_id = Variable(bits[1])
        self.section = bits[2] if len(bits) >= 3 else None

        self.varname = None
        if len(bits) > 3:
            if bits[3] != 'asvar':
                raise TemplateSyntaxError(
                    'Forth argument should be asvar," " got {}'.format(bits[3])
                )
            if len(bits) < 4:
                raise TemplateSyntaxError('Variable name not provided')
            else:
                self.varname = Variable(bits[4])

        self.nodelist = parser.parse(('end{}'.format(bits[0]),))
        parser.delete_first_token()

    def resolve_part(self, context):
        """Resolve field reference form context."""
        part = self.part_id.resolve(context)
        if isinstance(part, BoundField):
            part = part.field
        return part

    def render(self, context):  # noqa D102
        part = self.resolve_part(context)
        parts = context['form_parts']

        if self.section in parts[part]:
            # already rendered
            if self.varname is not None:
                varname = self.varname.resolve(context)
                context[varname] = parts[part][self.section]
                return ""
            else:
                return parts[part][self.section]

        # child parts
        children = (
            node for node in self.nodelist
            if isinstance(node, FormPartNode)
        )
        _render_parts(context, children)

        # render own content
        value = self.nodelist.render(context).strip()
        if self.varname is not None:
            context[self.varname.resolve(context)] = value
            return ''
        else:
            if not value:
                return ''
            return value


@register.tag('attrs')
class WidgetAttrsNode(Node):
    """
    Renders attrs for the html tag.

    <input{% attrs boundfield 'widget' default field.widget.attrs %}
        id="id_{{ bound_field.name }}"
        class="{% if bound_field.errors %}invalid{% endif %}"
    {% endattrs %}>
    """

    def __init__(self, parser, token):  # noqa D102
        bits = token.split_contents()

        if len(bits) < 3:
            raise TemplateSyntaxError(
                "%r accepts at least 2 arguments (bound_field,"
                " 'groupname'), got: {}".format(bits[0], ','.join(bits[1:]))
            )

        if len(bits) > 5:
            raise TemplateSyntaxError(
                "%r accepts at mast 4 arguments (bound_field, 'groupname'"
                " default attrs_dict ), got: {}".format(
                    bits[0], ','.join(bits[1:]))
            )

        if len(bits) > 3 and bits[3] != 'default':
            raise TemplateSyntaxError(
                "%r 3d argument should be 'default' (bound_field, 'groupname'"
                " default attrs_dict ), got: {}".format(
                    bits[0], ','.join(bits[1:]))
            )

        self.field = Variable(bits[1])
        self.group = Variable(bits[2])
        self.widget_attrs = Variable(bits[4]) if len(bits) >= 5 else None
        self.nodelist = parser.parse(('end{}'.format(bits[0])))
        parser.delete_first_token()

    def resolve_field(self, context):
        """Resolve field reference form context."""
        field = self.field.resolve(context)
        if isinstance(field, BoundField):
            field = field.field
        return field

    def render(self, context):  # noqa D102
        field = self.resolve_field(context)
        group = self.group.resolve(context)
        form_widget_attrs = context['form_widget_attrs']

        override = {}
        if group in form_widget_attrs[field]:
            override = form_widget_attrs[field][group]

        build_in_attrs, tag_content = {}, self.nodelist.render(context)
        for attr, _, value in ATTRS_RE.findall(tag_content):
            build_in_attrs[attr] = mark_safe(value) if value != '' else True

        widget_attrs = {}
        if self.widget_attrs is not None:
            widget_attrs = self.widget_attrs.resolve(context)

        result = build_in_attrs.copy()

        if 'class' in result and 'class' in widget_attrs:
            result['class'] += ' ' + widget_attrs.pop('class')
        result.update(widget_attrs)

        for attr, (value, action) in override.items():
            if action == 'override':
                result[attr] = value
            elif action == 'append':
                if attr in result:
                    result[attr] += " " + value
                else:
                    result[attr] = value

        return flatatt(result)


@register.tag('attr')
class WidgetAttrNode(Node):
    """The tag allows to add or override specific attribute in the rendered HTML.

    The first argumnent is the attribute group name, second is the
    attribute name.  The third optional flag shows to override (by
    default) or `append` the value.

    Usage::

        {% attr form.email 'widget' 'data-validate' %}email{% endattr %}
        {% attr form.email 'widget' 'class' append %}green{%  endattr %}
        {% attr form.email 'widget' 'required' %}True{%  endattr %}

    """

    def __init__(self, parser, token):  # noqa D102
        bits = token.split_contents()

        if len(bits) < 4:
            raise TemplateSyntaxError(
                "{} accepts at least 3 arguments (bound_field, 'groupname'"
                " 'attr_name'), got: {}".format(bits[0], ','.join(bits[1:]))
            )

        if len(bits) > 5:
            raise TemplateSyntaxError(
                "{} accepts at mast 4 arguments (bound_field, 'groupname'"
                " 'attr_name' action ), got: {}".format(
                    bits[0], ','.join(bits[1:]))
            )

        if len(bits) >= 5 and bits[4] not in ['append', 'override']:
            raise TemplateSyntaxError(
                "{} unknown action {}  should be 'append'"
                " of 'override'".format(bits[0], ','.join(bits[4]))
            )

        self.field = Variable(bits[1])
        self.group = Variable(bits[2])
        self.attr = bits[3]
        self.action = bits[4] if len(bits) >= 5 else 'override'
        self.nodelist = parser.parse(('end{}'.format(bits[0])))
        parser.delete_first_token()

    def resolve_field(self, context):
        """Resolve field reference form context."""
        field = self.field.resolve(context)
        if isinstance(field, BoundField):
            field = field.field
        return field

    def render(self, context):  # noqa D102
        field = self.resolve_field(context)
        group = self.group.resolve(context)
        form_widget_attrs = context['form_widget_attrs']
        value = self.nodelist.render(context)

        if group not in form_widget_attrs[field]:
            form_widget_attrs[field][group] = {}
        attrs = form_widget_attrs[field][group]

        if self.attr not in attrs or self.action == 'override':
            attrs[self.attr] = (value, self.action)
        else:
            old_value, old_action = attrs[self.attr]
            if old_action != 'override':
                attrs[self.attr] = (
                    '{} {}'.format(old_value, value),
                    self.action
                )



from . import modules as modules_registry


def modules(request):
    """Add current module and modules list to the template context."""
    if not hasattr(request, 'user'):
        raise ValueError('modules context processor requires "django.contrib.auth.context_processors.auth"'
                         'to be in TEMPLATE_CONTEXT_PROCESSORS in your settings file.')

    module = None

    if request.resolver_match:
        module = getattr(request.resolver_match.url_name, 'module', None)

    return {
        'modules': modules_registry.available_modules(request.user),
        'current_module': module,
    }



from __future__ import unicode_literals

from django.contrib import messages
from django.core.exceptions import PermissionDenied, ValidationError
from django.forms.models import modelform_factory
from django.http import Http404
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.html import format_html
from django.utils.http import urlquote
from django.utils.translation import ugettext as _

from ...base import Span


def _collect_elements(parent, container=None):
    if container is None:
        container = []

    if hasattr(parent, 'elements'):
        for element in parent.elements:
            _collect_elements(element, container=container)

    if isinstance(parent, Span):
        container.append(parent.field_name)

    return container


class ModelViewMixin(object):
    """Mixin for generic form views to play nice with `ModelViewSet`."""

    viewset = None
    layout = None
    form_widgets = None

    def __init__(self, *args, **kwargs):  # noqa D102
        super(ModelViewMixin, self).__init__(*args, **kwargs)
        if self.form_class is None and self.fields is None:
            if self.layout is not None:
                self.fields = _collect_elements(self.layout)
            else:
                self.fields = '__all__'

    def has_object_permission(self, request, obj):
        """Check object access permission.

        Subclasses should override it
        """
        raise NotImplementedError

    def get_queryset(self):
        """Return the list of items for this view.

        If view have no explicit `self.queryset`, tries too lookup to
        `viewflow.get_queryset`
        """
        if self.queryset is None and self.viewset is not None:
            if hasattr(self.viewset, 'get_queryset'):
                return self.viewset.get_queryset(self.request)
        return super(ModelViewMixin, self).get_queryset()

    def get_object(self):
        """Retrieve an object and check user permissions."""
        queryset = self.get_queryset()
        model = queryset.model
        pk = self.kwargs.get(self.pk_url_kwarg)
        if pk is not None:
            try:
                self.kwargs[self.pk_url_kwarg] = model._meta.pk.to_python(pk)
            except (ValidationError, ValueError):
                raise Http404

        obj = super(ModelViewMixin, self).get_object()
        if not self.has_object_permission(self.request, obj):
            raise PermissionDenied
        return obj

    def get_form_class(self):
        if self.form_class is None:
            if self.model is not None:
                model = self.model
            elif hasattr(self, 'object') and self.object is not None:
                model = self.object.__class__
            else:
                model = self.get_queryset().model
            return modelform_factory(model, fields=self.fields, widgets=self.form_widgets)
        return super(ModelViewMixin, self).get_form_class()

    def get_success_url(self):
        """Redirect back to the list view if no `success_url` is configured."""
        if self.success_url is None:
            opts = self.model._meta
            return reverse('{}:{}_list'.format(
                opts.app_label, opts.model_name)
            )
        return super(ModelViewMixin, self).get_success_url()

    def get_template_names(self):
        """
        List of templates for the view.

        If no `self.template_name` defined, uses::

             [<app_label>/<model_label>_<suffix>.html,
              <app_label>/<model_label>_form.html,
              'material/frontend/views/form.html']
        """
        if self.template_name is None:
            opts = self.model._meta
            return [
                '{}/{}{}.html'.format(
                    opts.app_label,
                    opts.model_name,
                    self.template_name_suffix),
                '{}/{}_form.html'.format(
                    opts.app_label,
                    opts.model_name),
                'material/frontend/views/form.html',
            ]

        return [self.template_name]

    def form_valid(self, *args, **kwargs):
        response = super(ModelViewMixin, self).form_valid(*args, **kwargs)
        self.message_user()
        return response

    def message_user(self):
        """Successful notification.

        Subclasses can override it.
        """


class MessageUserMixin(object):
    """User notification utilities django.messages framework."""

    def report(self, message, level=messages.INFO, fail_silently=True, **kwargs):
        """Construct message and notify the user."""
        opts = self.model._meta

        url = reverse('{}:{}_detail'.format(
            opts.app_label, opts.model_name), args=[self.object.pk])
        link = format_html('<a href="{}">{}</a>', urlquote(url), force_text(self.object))
        name = force_text(opts.verbose_name)

        options = {
            'link': link,
            'name': name
        }
        options.update(kwargs)
        message = format_html(_(message).format(**options))
        messages.add_message(self.request, messages.SUCCESS, message, fail_silently=True)

    def success(self, message, fail_silently=True, **kwargs):
        self.report(message, level=messages.SUCCESS, fail_silently=fail_silently, **kwargs)

    def error(self, message, fail_silently=True, **kwargs):
        self.report(message, level=messages.ERROR, fail_silently=fail_silently, **kwargs)



"""Django backport and compatibility utilities."""

import inspect
import functools

from django.utils import six
from django.utils.html import conditional_escape


__all__ = ('simple_tag', )


try:
    from django.template.library import Library  # NOQA
    simple_tag = Library.simple_tag
except ImportError:
    # django 1.8
    from django.template import Node
    from django.template.base import parse_bits

    class TagHelperNode(Node):
        """Base class for tag helper nodes such as SimpleNode and InclusionNode.

        Manages the positional and keyword arguments to be passed to
        the decorated function.
        """

        def __init__(self, func, takes_context, args, kwargs):  # noqa: D102
            self.func = func
            self.takes_context = takes_context
            self.args = args
            self.kwargs = kwargs

        def get_resolved_arguments(self, context):
            resolved_args = [var.resolve(context) for var in self.args]
            if self.takes_context:
                resolved_args = [context] + resolved_args
            resolved_kwargs = {
                k: v.resolve(context)
                for k, v in self.kwargs.items()
            }
            return resolved_args, resolved_kwargs

    class SimpleNode(TagHelperNode):
        def __init__(self, func, takes_context, args, kwargs, target_var):
            super(SimpleNode, self).__init__(func, takes_context, args, kwargs)
            self.target_var = target_var

        def render(self, context):
            resolved_args, resolved_kwargs = \
                self.get_resolved_arguments(context)
            output = self.func(*resolved_args, **resolved_kwargs)
            if self.target_var is not None:
                context[self.target_var] = output
                return ''
            if context.autoescape:
                output = conditional_escape(output)
            return output

    def getargspec(func):
        if six.PY2:
            return inspect.getargspec(func)

        sig = inspect.signature(func)
        args = [
            p.name for p in sig.parameters.values()
            if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
        ]
        varargs = [
            p.name for p in sig.parameters.values()
            if p.kind == inspect.Parameter.VAR_POSITIONAL
        ]
        varargs = varargs[0] if varargs else None
        varkw = [
            p.name for p in sig.parameters.values()
            if p.kind == inspect.Parameter.VAR_KEYWORD
        ]
        varkw = varkw[0] if varkw else None
        defaults = [
            p.default for p in sig.parameters.values()
            if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            if p.default is not p.empty
        ] or None
        return args, varargs, varkw, defaults

    def simple_tag(library, func=None, takes_context=None, name=None):
        """
        Register a callable as a compiled template tag.

        Example:
            @register.simple_tag
            def hello(*args, **kwargs):
                return 'world'

        Backport from django 1.9
        """
        def dec(func):
            params, varargs, varkw, defaults = getargspec(func)
            function_name = (
                name or getattr(func, '_decorated_function', func).__name__
            )

            @functools.wraps(func)
            def compile_func(parser, token):
                bits = token.split_contents()[1:]
                target_var = None
                if len(bits) >= 2 and bits[-2] == 'as':
                    target_var = bits[-1]
                    bits = bits[:-2]
                    args, kwargs = parse_bits(
                        parser, bits, params, varargs, varkw,
                        defaults, takes_context, function_name)
                return SimpleNode(
                    func, takes_context, args, kwargs, target_var)
            library.tag(function_name, compile_func)
            return func

        if func is None:
            # @register.simple_tag(...)
            return dec
        elif callable(func):
            # @register.simple_tag
            return dec(func)
        else:
            raise ValueError("Invalid arguments provided to simple_tag")


def context_flatten(context):
    result = {}
    # https://code.djangoproject.com/ticket/24765
    for dict_ in context.dicts:
        if hasattr(dict_, 'flatten'):
            dict_ = context_flatten(dict_)
        result.update(dict_)
    return result



"""Layout primitives definitions."""

import re
import warnings
from functools import partial
from django.template import TemplateDoesNotExist
from django.template.loader import get_template, select_template
from django.utils import six
from django.utils.encoding import smart_text


class LayoutNode(object):
    """Base class for self-rendered nodes."""

    span_columns = 1
    template_name = None

    def get_context_data(self, context):
        """Additional context data to render node template.

        Subclasses could override it.
        """
        return {}

    def get_template(self, context):
        """Determine template for the node.

        If `context` have the `template` variable it would be used as
        relative template name in the current template pack.

        Assumes `form_template_pack` context variable points to the
        current template pack

        """
        template_name = self.template_name
        if 'template' in context:
            template_name = context['template']
        return get_template("{}/{}".format(
            context['form_template_pack'], template_name))

    def render(self, context, **options):
        """
        Render node in the template.

        Compatible with django ``{% include %}`` tag.
        """
        context.push()
        try:
            for key, value in self.get_context_data(context).items():
                context[key] = value

            template = self.get_template(context)
            return template.render(context.flatten())
        finally:
            context.pop()


def _convert_to_field(elements):
    result = []
    for element in elements:
        if isinstance(element, six.string_types):
            result.append(Field(element))
        else:
            result.append(element)
    return result


def _camel_case_to_underscore(name):
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2',
                  re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)).lower()


def _get_field_template(template_pack, field):
    widget_templates = [
        '{}_{}.html'.format(
            cls.__module__.split('.', 1)[0],
            cls.__name__.lower())
        for cls in type(field.widget).mro()[:-2]
    ]

    field_templates = [
        '{}_{}/{}'.format(
            cls.__module__.split('.', 1)[0],
            cls.__name__.lower(),
            widget_template)
        for cls in type(field).mro()[:-2]
        for widget_template in widget_templates
    ]

    template_names = ["{}/fields/{}".format(template_pack, template_name)
                      for template_name in widget_templates + field_templates]

    return select_template(template_names)


class Layout(LayoutNode):
    """Form layout specification.

    Allows to set relative field sizes and positions.
    """

    template_name = 'layout/layout.html'

    def __init__(self, *elements):  # noqa: D102
        self.elements = _convert_to_field(elements)


class Fieldset(LayoutNode):
    """Group of relative elements in a form.

    :param label: Header for the fields group

    :param span_columns: Relative width of the fieldset
    """

    template_name = 'layout/fieldset.html'

    def __init__(self, label, *elements, **kwargs):  # noqa: D102
        self.label = label
        self.elements = _convert_to_field(elements)
        self.span_columns = kwargs.pop('span_columns', 1)


class Row(LayoutNode):
    """Place elements in a single line."""

    template_name = 'layout/row.html'

    def __init__(self, *elements, **kwargs):    # noqa: D102
        self.elements = _convert_to_field(elements)
        self.row_id = kwargs.pop('row_id', None)

    def __getattr__(self, name):
        _, container_size = name.split('_')
        container_size = int(container_size)

        def elements_iterator():
            elements_span = sum(
                element.span_columns for element in self.elements
            )
            if container_size % elements_span != 0:
                warnings.warn(
                    "Can't equally divide container {} for {} span elements"
                    .format(container_size, self.elements)
                )

            span_multiplier = container_size // elements_span
            for element in self.elements:
                yield element, element.span_columns * span_multiplier

        return elements_iterator


class Column(LayoutNode):
    """Place elements vertically stacked, one under another.

    :param span_columns: Relative width of the fieldset
    """

    template_name = 'layout/column.html'

    def __init__(self, *elements, **kwargs):  # noqa: D102
        self.elements = _convert_to_field(elements)
        self.span_columns = kwargs.pop('span_columns', 1)
        self.column_id = kwargs.pop('column_id', None)


class Span(object):
    """Wrapper for a field reference.

    There are  ``Span2``, ``Span3``, .., ``Span12`` shortcut classes.
    ``Layout`` autowraps string field references into Span(1, field_name)

    :param span_columns: relative field width
    :param field_name: field name in the form
    """

    template_name = 'layout/field.html'

    def __init__(self, span_columns, field_name):  # noqa: D102
        self.span_columns = span_columns
        self.field_name = field_name

    def render(self, context, **options):
        """
        Render field in the template.

        Compatible with django ``{% include %}`` tag.
        """
        template_pack = context['form_template_pack']
        form = context['form']
        bound_field = form[self.field_name]

        try:
            if 'template' in options:
                template = select_template([
                    "{}/{}".format(template_pack, options['template'])
                ])
            elif 'widget' in options:
                widget_templates = [
                    '{}/fields/{}_{}.html'.format(
                        template_pack,
                        cls.__module__.split('.', 1)[0],
                        cls.__name__.lower()
                    )
                    for cls in type(options['widget']).mro()[:-2]
                ]
                template = select_template(widget_templates)
            else:
                template = _get_field_template(
                    template_pack, bound_field.field)
        except TemplateDoesNotExist:
            # fallback to default field render
            warnings.warn("Unknown field and widget {} {}".format(
                bound_field.field.__class__,
                bound_field.field.widget.__class__))
            return smart_text(bound_field)
        else:
            hidden_initial = ''

            if bound_field.field.show_hidden_initial:
                hidden_initial = bound_field.as_hidden(only_initial=True)

            context.push()
            try:
                context['bound_field'] = bound_field
                context['field'] = bound_field.field
                context['hidden_initial'] = hidden_initial
                return template.render(context.flatten())
            finally:
                context.pop()

    def __str__(self):
        return 'Span{}({})'.format(self.span_columns, self.field_name)


Field = partial(Span, 1)
Span2 = partial(Span, 2)
Span3 = partial(Span, 3)
Span4 = partial(Span, 4)
Span5 = partial(Span, 5)
Span6 = partial(Span, 6)
Span7 = partial(Span, 7)
Span8 = partial(Span, 8)
Span9 = partial(Span, 9)
Span10 = partial(Span, 10)
Span11 = partial(Span, 11)
Span12 = partial(Span, 12)


def _collect_elements(element_cls, parent, container=None):
    if container is None:
        container = []

    if hasattr(parent, 'elements'):
        for element in parent.elements:
            _collect_elements(element_cls, element, container=container)

    if isinstance(parent, element_cls):
        container.append(parent)

    return container


class LayoutMixin(object):
    """Extracts from layout `fields` for django FormView.

    Shortcut allows don't specify ``fields`` parameter in the GenericView
    if you already  have layout object.

    .. code: python:

        class MyView(LayoutMixin, generics.CreateView):
            model = MyModel
            layout = Layout('field_1', 'field_2')
    """

    @property
    def fields(self):
        """Provide list of field for ModelForm."""
        fields = _collect_elements(Span, self.layout)
        return [field.field_name for field in fields]



from __future__ import unicode_literals

from django.contrib.auth import get_permission_codename
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import router
from django.db.models.deletion import Collector
from django.http import Http404
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _
from django.views import generic

from .mixins import MessageUserMixin


class DeleteModelView(MessageUserMixin, generic.DeleteView):
    """View for deleting an object and all linked by foreign key data."""

    viewset = None

    def has_object_permission(self, request, obj):
        """Object delete permission check.

        If view had a `viewset`, the `viewset.has_delete_permission` used.
        """
        if self.viewset is not None:
            return self.viewset.has_delete_permission(request, obj)

        # default lookup for the django permission
        opts = self.model._meta
        codename = get_permission_codename('delete', opts)
        delete_perm = '{}.{}'.format(opts.app_label, codename)
        if request.user.has_perm(delete_perm):
            return True
        return request.user.has_perm(delete_perm, obj=obj)

    def _get_deleted_objects(self):
        collector = Collector(using=router.db_for_write(self.object))
        collector.collect([self.object])
        return collector.data

    def get_context_data(self, **kwargs):
        """Extend view context data.

        `{{ deleted_objects }}` - list of related objects to delete
        """
        kwargs.setdefault('deleted_objects', self._get_deleted_objects())
        return super(DeleteModelView, self).get_context_data(**kwargs)

    def get_object(self):
        """Retrieve the object for delete.

        Check object delete permission at the same time.
        """
        queryset = self.get_queryset()
        model = queryset.model
        pk = self.kwargs.get(self.pk_url_kwarg)
        if pk is not None:
            try:
                self.kwargs[self.pk_url_kwarg] = model._meta.pk.to_python(pk)
            except (ValidationError, ValueError):
                raise Http404

        obj = super(DeleteModelView, self).get_object()
        if not self.has_object_permission(self.request, obj):
            raise PermissionDenied
        return obj

    def get_success_url(self):
        """Redirect back to the list view if no `success_url` is configured."""
        if self.success_url is None:
            opts = self.model._meta
            return reverse('{}:{}_list'.format(opts.app_label, opts.model_name))
        return super(DeleteModelView, self).get_success_url()

    def get_template_names(self):
        """
        List of templates for the view.

        If no `self.template_name` defined, uses::

             [<app_label>/<model_label>_delete.html,
              'material/frontend/views/confirm_delete.html']
        """
        if self.template_name is None:
            opts = self.model._meta
            return [
                '{}/{}{}.html'.format(opts.app_label, opts.model_name, self.template_name_suffix),
                'material/frontend/views/confirm_delete.html',
            ]

        return [self.template_name]

    def delete(self, request, *args, **kwargs):
        response = super(DeleteModelView, self).delete(request, *args, **kwargs)
        self.message_user()
        return response

    def message_user(self):
        self.success(_('The {name} "{link}"  was deleted successfully.'))



from __future__ import unicode_literals

import datetime
import decimal
import json

from collections import OrderedDict

from django.contrib.auth import get_permission_codename
from django.contrib.auth.decorators import login_required
from django.core.exceptions import FieldDoesNotExist
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.db.models.query import QuerySet
from django.forms.forms import pretty_name
from django.http import JsonResponse
from django.urls import reverse
from django.utils import formats, six, timezone
from django.utils.decorators import method_decorator
from django.utils.encoding import force_text
from django.utils.html import format_html
from django.views.generic import View
from django.views.generic.base import ContextMixin, TemplateResponseMixin

from .. import forms


def _get_attr_label(owner, attr_name):
    attr = getattr(owner, attr_name)
    if hasattr(attr, "short_description"):
        return attr.short_description
    elif isinstance(attr, property) and hasattr(attr, "fget"):
        if hasattr(attr.fget, "short_description"):
            return attr.fget.short_description
        else:
            return pretty_name(attr.fget.__name__)
    elif callable(attr):
        return "--" if attr.__name__ == "<lambda>" else pretty_name(attr.__name__)
    else:
        return attr_name


class ModelField(object):
    """Retrieve a field value from the model.

    Field verbose name would be use as a label.
    """

    def __init__(self, field):  # noqa D102
        self.field = field

    def get_value(self, obj):  # noqa D102
        return getattr(obj, self.field.name)

    @property
    def label(self):  # noqa D102
        try:
            return self.field.verbose_name
        except AttributeError:
            # field is likely a ForeignObjectRel
            return self.field.related_model._meta.verbose_name

    @property
    def orderable(self):  # noqa D102
        return True


class ModelAttr(object):
    """Retrieve attribute value from the model instance.

    If model attribute is a callable, to get the value it would be
    called without any arguments.
    """

    def __init__(self, model, name, label=None):  # noqa D102
        self.model = model
        self.name = name
        self._label = label

    def get_value(self, obj):  # noqa D102
        attr = getattr(obj, self.name)
        if callable(attr):
            return attr()
        return attr

    @property
    def label(self):  # noqa D102
        if self._label:
            return self._label
        return _get_attr_label(self.model, self.name)

    @property
    def orderable(self):  # noqa D102
        return False


class DataSourceAttr(object):
    """Retrieve attribute value from extenal data source.

    Data source attribute could be a property or callable.

    For a callable, to get the value it would be called with model
    instance.
    """

    def __init__(self, data_source, name):  # noqa D102
        self.data_source = data_source
        self.name = name

    def get_value(self, obj):  # noqa D102
        attr = getattr(self.data_source, self.name)
        if callable(attr):
            return attr(obj)
        return attr

    @property
    def label(self):
        """Retrieve the label for the data source attribute.

        Label could be specified in `.short_description` or the
        attribute name would be used.

        Example::

        class SampleDataSource(object):
            def description(self, model):
                return "Sample for {}".format(model)
            description.short_description = "Model Description"

        """
        return _get_attr_label(self.data_source, self.name)

    @property
    def orderable(self):  # noqa D102
        return False


class DataTableMixin(ContextMixin):
    """Mixing for list views with DataTable."""

    datatable_config = None
    datatable_default_config = {
        'processing': False,
        'serverSide': True,
        'ajax': {
            'url': '.',
        },
        'order': [],
        'ordering': True,
        'orderMulti': True,
        'info': False,
        'bFilter': False,
        'bAutoWidth': True,
        'bLengthChange': False,
        'oLanguage': {
            'oPaginate': {
                'sFirst': "",
                'sLast': "",
                'sNext': "&rang;",
                'sPrevious': "&lang;",
            }
        },
        'responsive': {
            'details': False
        }
    }
    list_display = ('__str__', )
    empty_value_display = ""
    ordering = None
    viewset = None
    paginate_by = 15

    def get_context_data(self, **kwargs):
        """Update view context.

        Include `datatable_config`, 'headers' and initial `data` to
        first page render.
        """
        context = super(DataTableMixin, self).get_context_data(**kwargs)
        context.update({
            'datatable_config': json.dumps(self.get_datatable_config()),
            'headers': self.get_headers_data(),
            'data': self.get_table_data(0, self.paginate_by),
        })

        return context

    def get_list_display(self):
        """Return list of columns to display."""
        return self.list_display

    def get_datatable_config(self):
        """Prepare datatable config."""
        config = self.datatable_default_config.copy()
        config['pageLength'] = self.paginate_by
        config['ajax']['url'] = self.request.path
        config['columns'] = self.get_columns_def()

        if self.ordering:
            datatable_ordering = []
            columns = self.get_list_display()

            for column in self.ordering:
                idx = columns.index(column.strip('-'))
                datatable_ordering.append([idx, 'asc' if column.startswith('-') else 'desc'])
            config['order'] = datatable_ordering

        if self.datatable_config is not None:
            config.update(self.datatable_config)
        return config

    def get_data_attr(self, attr_name):
        """Data getter for an attribute.

        Data could comes from the model field or external `data_source`
        method call.
        """
        opts = self.object_list.model._meta
        try:
            return ModelField(opts.get_field(attr_name))
        except FieldDoesNotExist:
            if attr_name == "__str__":
                return ModelAttr(self.object_list.model, attr_name, opts.verbose_name)
            else:
                data_sources = [self, self.viewset] if self.viewset is not None else [self]
                for data_source in data_sources:
                    if hasattr(data_source, attr_name):
                        return DataSourceAttr(data_source, attr_name)
            if hasattr(self.object_list.model, attr_name):
                return ModelAttr(self.object_list.model, attr_name)
        raise AttributeError("Unable to lookup '{}' on {}" .format(attr_name, self.object_list.model._meta.object_name))

    def get_columns_def(self):
        """Return columns definition for the datables js config."""
        return [
            {'data': field_name, 'orderable': self.get_data_attr(field_name).orderable}
            for field_name in self.get_list_display()
        ]

    def get_headers_data(self):
        """Readable column titles."""
        for field_name in self.get_list_display():
            attr = self.get_data_attr(field_name)
            yield field_name, attr.label

    def format_column(self, item, field_name, value):
        if value is None:
            return self.empty_value_display
        elif isinstance(value, datetime.datetime):
            return formats.localize(timezone.template_localtime(value))
        elif isinstance(value, (datetime.date, datetime.time)):
            return formats.localize(value)
        elif isinstance(value, six.integer_types + (decimal.Decimal, float)):
            return formats.number_format(value)
        elif isinstance(value, (list, tuple)):
            return ', '.join(force_text(v) for v in value)
        else:
            return force_text(value)

    def get_table_data(self, start, length):
        """Get a page for datatable."""
        for item in self.object_list[start:start + length]:
            columns = OrderedDict()
            for n, field_name in enumerate(self.get_list_display()):
                attr = self.get_data_attr(field_name)
                value = self.format_column(item, field_name, attr.get_value(item))
                columns[field_name] = value
            yield item, columns

    def total(self):
        """Total dataset size."""
        return self.object_list.count()

    def total_filtered(self):
        """Dataset size with filter applied."""
        return self.object_list.count()

    def get_ordering(self):
        """Return the field or fields to use for ordering the queryset."""
        if self.request_form.is_valid():
            ordering = []
            requested_order = self.request_form.cleaned_data['ordering']
            for spec in requested_order:
                column_num, column_dir = spec.get('column', 0), spec.get('dir', 'asc')

                try:
                    order = self.get_list_display()[int(column_num)]
                    if column_dir == 'asc':
                        order = '-' + order
                except (IndexError, TypeError):
                    """ Skip """
                else:
                    ordering.append(order)
            return ordering
        else:
            return self.ordering

    def get(self, request, *args, **kwargs):
        """Response with rendered html template."""
        context = self.get_context_data()
        return self.render_to_response(context)

    def get_json_data(self, request, *args, **kwargs):
        """Return `JSONResponse` with data for datatable."""
        if not self.request_form.is_valid():
            return JsonResponse({'error': self.request_form.errors})

        draw = self.request_form.cleaned_data['draw']
        start = self.request_form.cleaned_data['start']
        length = self.request_form.cleaned_data['length']

        result = []
        for item, columns_data in self.get_table_data(start, length):
            result.append(columns_data)

        data = {
            "draw": draw,
            "recordsTotal": self.total(),
            "recordsFiltered": self.total_filtered(),
            "data": result
        }

        return JsonResponse(data)

    def get_object_list(self):
        """Create prepared queryset for datatables view."""
        return self.get_queryset()

    def dispatch(self, request, *args, **kwargs):
        """Handle for browser HTTP and AJAX requests from datatables."""
        self.request_form = forms.DatatableRequestForm(request.GET, prefix='datatable')
        self.object_list = self.get_object_list()
        if 'HTTP_DATATABLE' in request.META:
            handler = self.get_json_data
        elif request.method.lower() in self.http_method_names:
            handler = getattr(
                self, request.method.lower(), self.http_method_not_allowed)
        else:
            handler = self.http_method_not_allowed
        return handler(request, *args, **kwargs)


class ListModelView(TemplateResponseMixin, DataTableMixin, View):
    """List view suitable to work with jQuery Datatables.

    The view responsive for handling GET/POST requests from the browser
    and AJAX from the datatable.

    :keyword model: Model class views

    :keyword queryset: Base views queryset

    :keyword paginate_by: Page size

    :keyword ordering: Default list order

    :keyword datatable_config: Datatable default config overrides

    :keyword list_display: List of fields to display

    :keyword list_display_links: List of fields form `list_display`
                                 linked to change view

    """

    model = None
    queryset = None
    template_name_suffix = '_list'
    list_display_links = ()

    def has_view_permission(self, request, obj=None):
        """Object view permission check.

        If view had a `viewset`, the `viewset.has_view_permission` used.
        """
        if self.viewset is not None:
            return self.viewset.has_view_permission(request, obj)

        # default lookup for the django permission
        opts = self.model._meta
        codename = get_permission_codename('view', opts)
        view_perm = '{}.{}'.format(opts.app_label, codename)
        if request.user.has_perm(view_perm):
            return True
        elif request.user.has_perm(view_perm, obj=obj):
            return True
        return self.has_change_permission(request, obj=obj)

    def has_change_permission(self, request, obj=None):
        """Object change permission check.

        If view had a `viewset`, the `viewset.has_change_permission` used.
        """
        if self.viewset is not None:
            return self.viewset.has_change_permission(request, obj)

        # default lookup for the django permission
        opts = self.model._meta
        codename = get_permission_codename('change', opts)
        change_perm = '{}.{}'.format(opts.app_label, codename)
        if request.user.has_perm(change_perm):
            return True
        return request.user.has_perm(change_perm, obj=obj)

    def has_add_permission(self, request):
        """Object add permission check.

        If view had a `viewset`, the `viewset.has_add_permission` used.
        """
        if self.viewset is not None:
            return self.viewset.has_add_permission(request)

        # default lookup for the django permission
        opts = self.model._meta
        codename = get_permission_codename('add', opts)
        return request.user.has_perm('{}.{}'.format(opts.app_label, codename))

    def get_template_names(self):
        """
        List of templates for the view.

        If no `self.template_name` defined, uses::

             [<app_label>/<model_label>_list.html,
              'material/frontend/views/list.html']
        """
        if self.template_name is None:
            opts = self.object_list.model._meta
            return [
                '{}/{}{}.html'.format(
                    opts.app_label,
                    opts.model_name,
                    self.template_name_suffix),
                'material/frontend/views/list.html',
            ]
        return [self.template_name]

    def get_list_display_links(self, list_display):
        """Return columns list that would be linked to the object details.

        If `self.list_display_links` is not set, the first column would be used.
        """
        if (self.list_display_links or
                self.list_display_links is None or
                not list_display):
            return self.list_display_links
        else:
            # Use only the first item in list_display as link
            return list(list_display)[:1]

    def get_queryset(self):
        """Retrieve the view queryset."""
        if self.queryset is not None:
            queryset = self.queryset
            if isinstance(queryset, QuerySet):
                queryset = queryset.all()
        elif self.viewset is not None and hasattr(self.viewset, 'get_queryset'):
            queryset = self.viewset.get_queryset(self.request)
        elif self.model is not None:
            queryset = self.model._default_manager.all()
        else:
            raise ImproperlyConfigured(
                "%(cls)s is missing a QuerySet. Define "
                "%(cls)s.model, %(cls)s.queryset, or override "
                "%(cls)s.get_queryset()." % {
                    'cls': self.__class__.__name__
                })
        ordering = self.get_ordering()
        if ordering:
            if isinstance(ordering, six.string_types):
                ordering = (ordering,)
            queryset = queryset.order_by(*ordering)
        return queryset

    def format_column(self, item, field_name, value):
        if isinstance(value, bool):
            return format_html('<i class="material-icons">{}</i>'.format(
                'check' if value else 'close'
            ))
        else:
            formatted = super(ListModelView, self).format_column(item, field_name, value)
            if field_name in self.get_list_display_links(self.get_list_display()):
                formatted = format_html('<a href="{}">{}</a>', self.get_item_url(item), formatted)
            return formatted

    def get_item_url(self, item):
        """Link to object detail to `list_display_links` columns."""
        opts = self.model._meta
        return reverse(
            '{}:{}_detail'.format(opts.app_label, opts.model_name),
            args=[item.pk])

    def get_context_data(self, **kwargs):
        """Additional context data for list view.

        :keyword add_url: Link to the add view
        """
        opts = self.model._meta

        if self.has_add_permission(self.request):
            kwargs['add_url'] = reverse('{}:{}_add'.format(opts.app_label, opts.model_name))

        return super(ListModelView, self).get_context_data(**kwargs)

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        """Handle for browser HTTP and AJAX requests from datatables."""
        if not self.has_view_permission(self.request):
            raise PermissionDenied

        return super(ListModelView, self).dispatch(request, *args, **kwargs)



class Registry(object):
    """Frontend module registry."""

    def __init__(self):  # noqa D102
        self._registry = {}

    def modules(self):
        """List of modules in the frontend according to it's order."""
        return sorted([module for module in self._registry.values()],
                      key=lambda module: (module.order, module.label))

    def installed_modules(self):
        """List of installed modules.

        A frontend module itself determines if it installed or not.
        If the module instance have no `installed` attribute, the
        module considered installed.
        """
        return [module for module in self.modules()
                if getattr(module, 'installed', True)]

    def available_modules(self, user):
        """List of modules permitted for the user."""
        return [module for module in self.installed_modules()
                if module.has_perm(user)]

    def get_module(self, module_label):
        """Get module by label."""
        return self._registry.get(module_label, None)

    def register(self, module):
        """Register a module.

        Can be called on the package level.
        """
        self._registry[module.label] = module

    @property
    def urls(self):
        """Collected list of all modules url.

        Even uninstalled modules urls returned here. The module url
        config have the responsibility to check the module installed
        state.
        """
        patterns = []
        for module in self.modules():
            patterns.append(module.urls)
        return patterns


modules = Registry()



from __future__ import division

import math
import re
from collections import OrderedDict

from django.db.models.query import QuerySet
from django.forms.forms import BoundField
from django.template import Library
from django.template.base import (
    Node, TemplateSyntaxError, Variable, token_kwargs
)
from django.utils import formats
from django.utils.encoding import force_text

from ..base import Field
from ..widgets import SelectDateWidget
from .material_form import FormPartNode, WidgetAttrNode, _render_parts


register = Library()


@register.tag('render')
class FormRenderNode(Node):
    """Sugar for element in template rendering."""

    def __init__(self, parser, token):  # noqa D102
        bits = token.split_contents()

        if len(bits) == 0:
            raise TemplateSyntaxError(
                "%r received invalid args, expected one element for render."
                " Got: %r".format(bits[0], bits[1:])
            )

        remaining_bits = bits[2:]

        self.kwargs = token_kwargs(remaining_bits, parser)

        if remaining_bits:
            raise TemplateSyntaxError("%r received an invalid token: %r" %
                                      (bits[0], remaining_bits[0]))

        for key in self.kwargs:
            if key not in ('template', 'widget'):
                raise TemplateSyntaxError("%r received an invalid key: %r" %
                                          (bits[0], key))

            self.kwargs[key] = self.kwargs[key]

        self.nodelist = parser.parse(('end{}'.format(bits[0]),))
        parser.delete_first_token()

        self.element = Variable(bits[1])

    def render(self, context):  # noqa D102
        element = self.element.resolve(context)

        options = {}
        for key, value in self.kwargs.items():
            options[key] = value.resolve(context)

        # render inner parts
        children = (
            node for node in self.nodelist
            if isinstance(node, FormPartNode)
        )
        _render_parts(context, children)

        attrs = (
            node for node in self.nodelist
            if isinstance(node, WidgetAttrNode)
        )
        for attr in attrs:
            attr.render(context)

        # render element
        if isinstance(element, BoundField):
            return Field(element.name).render(context, **options)
        elif hasattr(element, 'render'):
            with context.push(parent=element):
                return element.render(context, **options)
        else:
            raise TemplateSyntaxError(
                "form_render can't render %r".format(element)
            )


@register.filter
def multiwidget_value(bound_field, pos):
    """Subwidget value for MultiWidget."""
    value = bound_field.value()
    if not isinstance(value, (list, tuple)):
        value = bound_field.field.widget.decompress(value)
    return value[pos]


@register.filter
def have_default_choice(field):
    """Handle special case for SelectMultiple widget."""
    return [
        choice
        for choice, _ in field.widget.choices
        if choice is None or choice == ""
    ]


@register.filter
def jquery_datepicker_format(field):
    """Convert django input format to jquery datepicket format."""
    input_format = field.input_formats[0]

    # %a, %A, %z, %f %Z %j %U %W %c %x %X unsupported

    subst = {
        '%d': 'd',    # Day of the month as a zero-padded decimal number
        '%b': 'M',    # Month as locale's abbreviated name
        '%B': 'F',    # Month as locale's full name
        '%m': 'm',    # Month as a zero-padded decimal number
        '%y': 'y',    # Year without century as a zero-padded decimal number
        '%Y': 'Y',    # Year with century as a decimal number
        '%H': 'H',    # Hour (24-hour clock) as a zero-padded decimal number
        '%I': 'h',    # Hour (12-hour clock) as a zero-padded decimal number
        '%p': 'a',    # Locale's equivalent of either AM or PM
        '%M': 'i',    # Minute as a zero-padded decimal number
        '%S': 's',    # Second as a zero-padded decimal number
        '%%': '%'     # A literal '%' character
    }

    return re.sub('|'.join(re.escape(key) for key in subst.keys()),
                  lambda k: subst[k.group(0)], input_format)


@register.filter
def datepicker_value(value, date_format):
    """Return localized date value."""
    return formats.localize_input(value, date_format)


@register.filter('force_text')
def force_text_impl(value):
    """Coerce widget value to text."""
    return force_text(value)


@register.filter
def split_choices_by_columns(choices, columns):
    """Split CheckboxSelectMultiple values into columns."""
    columns = int(columns)
    col_span = 12 // columns
    per_column = int(math.ceil(len(choices) / columns))
    choices = [tuple(choice) + (i,) for i, choice in enumerate(choices)]
    return [
        (col_span, choices[i:i + per_column])
        for i in range(0, len(choices), per_column)
    ]


@register.filter
def select_date_widget_wrapper(bound_field):
    """Wrap SelectDateWidget into django-material internal wrapper."""
    class Wrapper(object):
        def __init__(self, bound_field):
            self.bound_field = bound_field

        @property
        def selects(self):
            widget = SelectDateWidget(self.bound_field.field.widget)
            for data in widget.selects_data(self.bound_field.value()):
                yield data

    return Wrapper(bound_field)


@register.filter
def is_initial_file(value):
    """Check for initial value of FileFile."""
    return bool(value and getattr(value, 'url', False))


@register.filter
def is_null_boolean_selected(bound_field, value):
    """Return NullBooleanField state."""
    BOOL_VALUES = {True: '2', False: '3', '2': '2', '3': '3'}
    try:
        current_value = BOOL_VALUES[bound_field.value()]
    except KeyError:
        current_value = '1'
    return value == current_value


@register.filter
def select_options(bound_field):
    """
    Return list of (group_name, option_label, option_value, selected).

    If group_name is None - option is not belongs to group
    """
    selected = bound_field.value()
    if not isinstance(selected, (list, tuple, QuerySet)):
        selected = [selected]
    selected = set(force_text(v) for v in selected)

    groups = OrderedDict()
    for option in bound_field.field.widget.choices:
        option_value, option_label = option[0], option[1]
        if isinstance(option_label, (list, tuple)):
            if option_value not in groups:
                groups[option_value] = []
            for value, label in option_label:
                if value is None:
                    value = ''
                value = force_text(value)
                groups[option_value].append((label, value, value in selected))
        else:
            if None not in groups:
                groups[None] = []
            if option_value is None:
                option_value = ''
            value = force_text(option_value)
            groups[None].append(
                (option_label, option_value, value in selected)
            )

    return groups.items()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('frontend', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='module',
            options={'verbose_name_plural': 'modules', 'verbose_name': 'module'},
        ),
        migrations.AlterField(
            model_name='module',
            name='installed',
            field=models.BooleanField(default=True, verbose_name='installed'),
        ),
        migrations.AlterField(
            model_name='module',
            name='label',
            field=models.SlugField(verbose_name='label'),
        ),
    ]



import warnings
from importlib import import_module

from django.apps import AppConfig, apps
from django.db.models.signals import post_migrate
from django.template import Template, TemplateDoesNotExist
from django.template.loader import get_template, select_template
from django.urls import reverse
from django.utils.module_loading import module_has_submodule
from django.utils.six.moves import input
from django.utils.translation import ugettext_lazy as _

from .registry import modules as modules_registry
from .urlconf import ModuleURLResolver


class ModuleMixin(object):
    """Extension for the django AppConfig. Makes django app pluggable at runtime.

    The application have to have <app_module>/urls.py file, with
    a single no-parametrized url with name='index', ex::

        urlpatterns = [
            url('^$', generic.TemplateView.as_view(template_name="sales/index.html"), name="index"),
        ]

    All AppConfigs urls will be included into material.frontend.urls automatically under /<app_label>/ prefix
    The AppConfig.label, used for the urls namespace.

    The menu.html sample::

        <ul>
            <li><a href="{% url 'sales:index' %}">Dashboard</a></li>
            <li><a href="{% url 'sales:customers' %}">Customers</a></li>
            {% if perms.sales.can_add_lead %}<li><a href="{% url 'sales:leads' %}">Leads</a></li>{% endif %}
        </ul>

    In all application templates, the current application config
    instance would be available as `current_module` template variable

    :keyword order: The relative module order priority. Modules in the
                    site menu would be listed according its priorities.

    :keyword icon: The module icon.

    Example::

        class Sales(ModuleMixin, AppConfig):
            name = 'sales'
            icon = '<i class="material-icons">call</i>'

    """

    order = 10
    icon = '<i class="material-icons">receipt</i>'

    @property
    def verbose_name(self):
        """Module name."""
        return self.label.title()

    @property
    def installed(self):
        """Check the module installation state.

        Default implementation store installed state in the database
        `frontend_dbmodule` table.
        """
        from .models import Module as DbModule
        return DbModule.objects.installed(self.label)

    def description(self):
        """Module description.

        By default taken from the module docstring.
        """
        return (self.__doc__ or "").strip()

    def has_perm(self, user):
        """Check is user have permission to access to the module."""
        return True

    def get_urls(self):  # noqa D102
        if module_has_submodule(self.module, 'urls'):
            urls_module_name = '%s.%s' % (self.name, 'urls')
            urls_module = import_module(urls_module_name)
            if hasattr(urls_module, 'urlpatterns'):
                return urls_module.urlpatterns

        warnings.warn('Module {} have not urls.py submodule or `urlpatterns` in it'.format(self.label))
        return []

    @property
    def urls(self):
        """Module url config.

        By default it would be loaded from '<app>/urls.py'
        """
        base_url = r'^{}/'.format(self.label)
        return ModuleURLResolver(base_url, self.get_urls(), module=self, app_name=self.label, namespace=self.label)

    def index_url(self):
        """Entry url for a module."""
        return reverse('{}:index'.format(self.label))

    def menu(self):
        """Load module menu template.

        Template should be located in `<app_label>/menu.html`

        If no template exists, no exception raised.

        Intended to use with {% include %} template tag::

            {% include module.menu %}
        """
        try:
            return get_template('{}/menu.html'.format(self.label))
        except TemplateDoesNotExist:
            return Template('')

    def base_template(self):
        """Base template for a module.

        If  <app_label>/base_module.html exists it would be used.
        The default is 'material/frontend/base_module.html'

        Intended to use in modules generic templates. Ex::

            {% extends current_module.base_template %}
        """
        return select_template([
            '{}/base_module.html'.format(self.label),
            'material/frontend/base_module.html'
        ])


class MaterialFrontendConfig(AppConfig):
    """Default config for Material Frontend."""

    name = 'material.frontend'
    verbose_name = _('Frontend')
    icon = '<i class="material-icons">view_module</i>'

    def ready(self):
        """Register all available modules."""
        for app_config in apps.get_app_configs():
            if isinstance(app_config, ModuleMixin):
                modules_registry.register(app_config)
        post_migrate.connect(update_modules, sender=self)


def update_modules(app_config, verbosity=2, interactive=True, **kwargs):
    """Sync installed modules and database records."""
    from .models import Module as DbModule

    for module in modules_registry.modules():
        _, created = DbModule.objects.get_or_create(label=module.label)
        if created and verbosity >= 2:
            print('Adding module {}'.format(module.label))

    stale_modules = DbModule.objects.exclude(
        label__in=[module.label for module in modules_registry.modules()])

    if stale_modules.exists():
        if interactive:
            ok_to_delete = input(
                "The following modules are stale and need to be deleted:\n {}\n"
                "Are you sure you want to delete these modules entries?\n\n"
                "Type 'yes' to continue, or 'no' to cancel: ".format(
                    '    %s'.join(module.label for module in stale_modules)
                ))
        else:
            ok_to_delete = 'yes'

        if ok_to_delete == 'yes':
            stale_modules.delete()
            print("Stale modules deleted.")
        else:
            if verbosity >= 2 or interactive:
                print("Stale modules remain.")



import datetime

from django.conf import settings
from django.utils import formats, six
from django.forms.widgets import Widget
from django.utils.encoding import force_str


class SelectDateWidget(Widget):
    """Wrapper around django.widgets.SelectDateWidget.

    Provides api suitable for template-based rendering
    """

    def __init__(self, widget):  # noqa: D102
        self.widget = widget

    @property
    def date_re(self):
        """Backward compatible date regexp source."""
        if hasattr(self.widget, 'date_re'):
            return self.widget.date_re
        else:
            # django 1.8
            from django.forms.widgets.extra import DATE_RE
            return DATE_RE

    def split_value(self, value):
        """Bit magic for widget value splitting into date components."""
        try:
            year_val, month_val, day_val = value.year, value.month, value.day
        except AttributeError:
            year_val = month_val = day_val = None
            if isinstance(value, six.string_types):
                if settings.USE_L10N:
                    try:
                        input_format = formats.get_format(
                            'DATE_INPUT_FORMATS')[0]
                        v = datetime.datetime.strptime(
                            force_str(value), input_format)
                        year_val, month_val, day_val = v.year, v.month, v.day
                    except ValueError:
                        pass
                else:
                    match = self.re_date.match(value)
                    if match:
                        year_val, month_val, day_val = [
                            int(v) for v in match.groups()
                        ]

        return year_val, month_val, day_val

    def parse_date_fmt(self):
        """List of year/month/day in order according to `DATE_FORMAT`."""
        fmt = formats.get_format('DATE_FORMAT')
        escaped = False
        for char in fmt:
            if escaped:
                escaped = False
            elif char == '\\':
                escaped = True
            elif char in 'Yy':
                yield 'year'
            elif char in 'bEFMmNn':
                yield 'month'
            elif char in 'dj':
                yield 'day'

    def none_choice(self, none_value):
        """Value for the empty select option."""
        return [] if self.widget.is_required else [none_value]

    def selects_data(self, value):
        """Content for the rendering select widgets."""
        year_val, month_val, day_val = self.split_value(value)

        year_choices = (
            self.none_choice(self.widget.year_none_value) +
            [(i, i) for i in self.widget.years]
        )

        month_choices = (
            self.none_choice(self.widget.month_none_value) +
            list(six.iteritems(self.widget.months))
        )

        day_choices = (
            self.none_choice(self.widget.day_none_value) +
            [(i, i) for i in range(1, 32)]
        )

        data = {
            'year': {
                'type': 'year',
                'value': year_val,
                'choices': year_choices
            },
            'month': {
                'type': 'month',
                'value': month_val,
                'choices': month_choices
            },
            'day': {
                'type': 'day',
                'value': day_val,
                'choices': day_choices
            },
        }

        for field in self.parse_date_fmt():
            yield data[field]



from setuptools import setup

try:
    long_description = open('README.rst').read()
except IOError:
    long_description = ''

setup(
    name='django-material',
    version='1.2.0',
    description='Material design for django forms and admin',
    license='BSD',
    author='Mikhail Podgurskiy',
    author_email='kmmbvnr@gmail.com',
    url='http://github.com/viewflow/django-material',
    keywords="django",
    packages=[
        'material',
        'material.templatetags',
        'material.theme',
        'material.theme.amber',
        'material.theme.bluegrey',
        'material.theme.cyan',
        'material.theme.deeppurple',
        'material.theme.indigo',
        'material.theme.lightgreen',
        'material.theme.orange',
        'material.theme.purple',
        'material.theme.teal',
        'material.theme.blue',
        'material.theme.brown',
        'material.theme.deeporange',
        'material.theme.green',
        'material.theme.lightblue',
        'material.theme.lime',
        'material.theme.pink',
        'material.theme.red',
        'material.theme.yellow',
        'material.frontend',
        'material.frontend.management',
        'material.frontend.management.commands',
        'material.frontend.migrations',
        'material.frontend.templatetags',
        'material.frontend.views',
        'material.admin',
    ],
    include_package_data=True,
    zip_safe=False,
    platforms=['any'],
    install_requires=[
    ],
    classifiers=[
        'Framework :: Django',
        "Framework :: Django :: 1.11",
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
        'Topic :: Software Development :: User Interfaces',
    ],
)



from django.contrib import admin
from . import models


@admin.register(models.Module)
class ModuleAdmin(admin.ModelAdmin):  # noqa D102
    actions = None
    icon = '<i class="material-icons">view_module</i>'
    list_display = ['label', 'installed']
    readonly_fields = ['label']

    def has_add_permission(self, request):
        """Module added automatically during the database migration."""
        return False

    def has_delete_permission(self, request, obj=None):
        """Module deletion is no allowed."""
        return False



from __future__ import unicode_literals

try:
    from urllib.parse import quote
except:
    from urllib import quote

from django.urls import Resolver404
from django.http.request import QueryDict
from django.utils import six


class ModuleMatchName(str):
    """Dump str wrapper.

    Just to keep module reference over django url resolve calling
    hierarchy.
    """


try:
    from django.urls import URLResolver
    from django.urls.resolvers import RegexPattern

    class ModuleURLResolver(URLResolver):
        def __init__(self, regex, urlconf_name, default_kwargs=None, app_name=None, namespace=None, module=None):  # noqa D102
            self._module = module
            if app_name is None and namespace is not None:
                app_name = namespace
            pattern = RegexPattern(regex, is_endpoint=False)
            super(ModuleURLResolver, self).__init__(
                pattern,
                urlconf_name,
                default_kwargs,
                app_name=app_name,
                namespace=namespace,
            )

        def resolve(self, *args, **kwargs):  # noqa D102
            result = super(ModuleURLResolver, self).resolve(*args, **kwargs)

            if result and not getattr(self._module, 'installed', True):
                raise Resolver404({'message': 'Module not installed'})

            result.url_name = ModuleMatchName(result.url_name)
            result.url_name.module = self._module

            return result
except ImportError:
    # django 1.11
    from django.urls import RegexURLResolver

    class ModuleURLResolver(RegexURLResolver):
        """Module URL Resolver.

        A wrapper around RegexURLResolver that check the module installed
        state. And allows access to the resolved current module at runtime.

        Django reads url config once at the start. Installation and
        uninstallation the module at runtime don't produce change in the
        django url-conf.

        Url access check happens at the resolve time.
        """

        def __init__(self, *args, **kwargs):  # noqa D102
            self._module = kwargs.pop('module')
            super(ModuleURLResolver, self).__init__(*args, **kwargs)

        def resolve(self, *args, **kwargs):  # noqa D102
            result = super(ModuleURLResolver, self).resolve(*args, **kwargs)

            if result and not getattr(self._module, 'installed', True):
                raise Resolver404({'message': 'Module not installed'})

            result.url_name = ModuleMatchName(result.url_name)
            result.url_name.module = self._module

            return result


def frontend_url(request, url=None, back_link=None, absolute=True):
    """Construct an url for a frontend view.

    :keyword back: type of the back link to be added to the query string
                   - here: link to the current request page
                   - here_if_none: add link only if there is no `back` parameter
    :keyword absolute: construct absolute url, including host name

    namespace = self.ns_map[task.process.flow_class]
    return frontend_url(
        self.request,
        flow_url(namespace, task, 'index', user=request.user),
        back='here')
    """
    params = QueryDict(mutable=True)
    for key, value in six.iterlists(request.GET):
        if not key.startswith('datatable-') and key != '_':
            params.setlist(key, value)

    if back_link == 'here_if_none' and 'back' in params:
        # Do nothing
        pass
    elif back_link is not None:
        if params:
            back = "{}?{}".format(quote(request.path), quote(params.urlencode()))
        else:
            back = "{}".format(quote(request.path))
        params['back'] = back

    if url is not None:
        location = '{}?{}'.format(url, params.urlencode())
        return request.build_absolute_uri(location) if absolute else location
    else:
        return params.urlencode()



from django.db import models
from django.template import Library, TemplateSyntaxError

from ..urlconf import frontend_url


register = Library()


@register.filter
def frontend_urlname(model, url_type):
    """
    Retrieve URL for a model.

    Suitable for use with {% url %} template tag

    Example::

        {% url view.model|frontend_urlname:'list' %}
        {% url view.model|frontend_urlname:'detail' object.pk %}

    """
    return '{}:{}_{}'.format(model._meta.app_label, model._meta.model_name, url_type)


@register.filter
def verbose_name(obj):
    """Return model verbose name."""
    if isinstance(obj, models.Model):
        type(obj)._meta.verbose_name
    return obj._meta.verbose_name


@register.filter
def verbose_name_plural(obj):
    """Return model verbose name in plural mode."""
    if isinstance(obj, models.Model):
        type(obj)._meta.verbose_name_plural
    return obj._meta.verbose_name_plural


@register.filter
def query_back(request, back_link=None):
    """
    Set the `back` url GET parameter.

    Usage::

        <a href="{{ url }}?{{ request|query_back:"here" }}>
        <a href="{{ url }}?{{ request|query_back:"here_if_none" }}>
    """
    if back_link is not None and back_link not in ['here', 'here_if_none']:
        raise TemplateSyntaxError(
            "query_back tag accepts `here`, or `here_if_none` as parameter. Got {}".format(back_link))

    return frontend_url(request, back_link=back_link)
